# moshmosh?
# +template-python
from moshmosh.extension import Extension
from moshmosh.ast_compat import ast
from moshmosh.extensions.lazy_import.runtime import LazyModule

runtime_lazy_mod = "_moshmosh_lazy_module"


class LazyImportVisitor(ast.NodeTransformer):
    def __init__(self, act):
        self.act = act

    def visit_Import(self, imp: ast.Import):
        if imp.lineno not in self.act:
            return imp

        @quote
        def f(lazy_module, name, asname, names_var):
            for name, asname, in names_var:
                lazy_module(globals(), name, asname)

        names_var = ast.Constant(tuple((name.name, name.asname) for name in imp.names))
        name = ast.Name(".name")
        asname = ast.Name(".asname")
        return f(ast.Name(runtime_lazy_mod, ast.Load()), name, asname, names_var)

    def visit_ImportFrom(self, imp: ast.ImportFrom):
        if imp.lineno not in self.act or imp.names[0].name == '*':
            return imp
        if imp.module is None:
            from_mod = ast.Name(".from_mod")
            @quote
            def mk_from_mod(from_mod_n, level):
                from_mod_n = '.'.join(__name__.split('.')[:-level])
            mk_from_mod = mk_from_mod(from_mod, ast.Constant(imp.level))
        else:
            mk_from_mod = []
            from_mod = ast.Constant(imp.module)

        names_var = ast.Constant(tuple((name.name, name.asname) for name in imp.names))
        name = ast.Name(".name")
        asname = ast.Name(".asname")
        @quote
        def f(lazy_module, from_mod, name, asname, names_var):
            for name, asname, in names_var:
                lazy_module(globals(), name, asname, from_mod)
        stmts = f(
            ast.Name(runtime_lazy_mod, ast.Load()),
            from_mod,
            name,
            asname,
            names_var
        )
        mk_from_mod.extend(stmts)
        return stmts


class LazyImport(Extension):
    identifier = "lazy-import"

    def __init__(self):
        self.visitor = LazyImportVisitor(self.activation)

    def rewrite_ast(self, node):
        node = ast.fix_missing_locations(self.visitor.visit(node))
        # from astpretty import pprint
        # pprint(node)
        return node

    def pre_rewrite_src(self, io):
        io.write("from {} import {} as {}\n".format(__name__, LazyModule.__name__, runtime_lazy_mod))



import ast as _ast
from types import ModuleType
from sys import version_info

from numbers import Number


class SupertypeMeta(type):
    def __instancecheck__(self, other):
        return isinstance(other, self._sup_cls)


class ConsistentConstant(metaclass=SupertypeMeta):
    _sup_cls = (_ast.Num, _ast.NameConstant, _ast.Str)

    def __new__(self, i):
        if isinstance(i, Number):
            return ast.Num(i)
        if isinstance(i, str):
            return ast.Str(i)
        if isinstance(i, tuple) and version_info < (3, 6):
            return ast.Tuple(elts=list(map(ConsistentConstant, i)), ctx=ast.Load())
        return ast.NameConstant(i)

if version_info < (3, 7):
    ast = ModuleType("ast", _ast.__doc__)

    def make_new_init(supercls, fields_):
        def init(self, *args, **kwargs):
            supercls.__init__(self)
            fields = iter(fields_)
            for arg in args:
                field = next(fields)
                setattr(self, field, arg)
            for field in fields:
                v = kwargs.get(field)
                if v is not None:
                    setattr(self, field, v)

        return init

    for k, v in _ast.__dict__.items():
        if isinstance(v, type) and issubclass(v, _ast.AST):
            ns = {"__init__": make_new_init(v, v._fields), "_sup_cls": v}
            v = SupertypeMeta(k, (v, ), ns)

        setattr(ast, k, v)

    if not hasattr(_ast, "Constant"):

        class Constant(ConsistentConstant):
            pass

        ast.Constant = Constant
        def get_constant(n: Constant):
            if isinstance(n, ast.Num):
                return n.n
            elif isinstance(n, ast.Str):
                return n.s
            return n.value
    else:
        def get_constant(n: ast.Constant):
            return n.value

    if not hasattr(_ast, "Starred"):

        class Starred:
            def __init__(self, *args, **kwargs):
                raise NotImplementedError

        ast.Starred = Starred

    if not hasattr(_ast, "AnnAssign"):

        class AnnAssign:
            def __init__(self, *args, **kwargs):
                raise NotImplementedError

        ast.AnnAssign = AnnAssign
else:
    ast = _ast
    def get_constant(n: ast.Constant):
        return n.value


from moshmosh.extension import Extension
from moshmosh.ast_compat import ast


class PipelineVisitor(ast.NodeTransformer):
    """
    `a | f -> f(a)`, recursively
    """
    def __init__(self, activation):
        self.activation = activation

    def visit_BinOp(self, n: ast.BinOp):
        if n.lineno in self.activation and isinstance(n.op, ast.BitOr):
            return ast.Call(
                self.visit(n.right),
                [self.visit(n.left)],
                [],
                lineno=n.lineno,
                col_offset=n.col_offset
            )
        return self.generic_visit(n)

class Pipeline(Extension):
    identifier = "pipeline"
    def __init__(self):
        self.visitor = PipelineVisitor(self.activation)

    def rewrite_ast(self, node):
        return self.visitor.visit(node)



from moshmosh.extensions.pattern_matching.runtime import ListView
v = ListView([1, 2, 3, 5], range(1, 4))
v.sort(reverse=True)
assert v == [5, 3, 2]

v.sort(reverse=False)
assert v == [2, 3, 5]

v.sort(reverse=False, key=lambda x: -x)

assert v == [5, 3, 2]

assert isinstance(v, list)


from io import StringIO
from moshmosh.ast_compat import ast
from moshmosh.ast_compat import ConsistentConstant
from moshmosh.extensions.pattern_matching.core import *
from moshmosh.extension import Extension, Activation
from moshmosh.extensions.pattern_matching.runtime import NotExhaustive
from moshmosh.ctx_fix import ExprContextFixer


class SyntacticPatternBinding:
    def __init__(self, case_comp: CaseCompilation):
        self.case_comp = case_comp

    def visit_Name(self, n: ast.Name):
        if n.id == "_":
            return self.case_comp.wildcard()
        return self.case_comp.capture(Symbol(n.id, n.lineno, n.col_offset))

    def visit_Call(self, n: ast.Call):
        if n.keywords:
            raise NotImplementedError(n)

        if isinstance(n.func, ast.Name):
            if n.func.id == 'pin' and len(
                n.args) == 1:
                return self.case_comp.pin(Expr(n.args[0]))
            if n.func.id == 'isinstance':
                if n.args == 1:
                    expr = Expr(n.args[0])
                else:
                    expr = Expr(ast.Tuple(n.args, ast.Load()))
                return self.case_comp.instance_of(expr)

        return self.case_comp.recog2(
            Expr(n.func), [self.visit(elt) for elt in n.args])

    def visit_BoolOp(self, n: ast.BoolOp):

        if isinstance(n.op, ast.And):
            cases = list(map(self.visit, n.values))
            return self.case_comp.intersect(cases)
        if isinstance(n.op, ast.Or):
            cases = list(map(self.visit, n.values))
            return self.case_comp.alternative(cases)

        raise NotImplementedError(n)

    def _visit_seq(self, type, n):
        def find_star(elts):
            for i, elt in enumerate(elts):
                if isinstance(elt, ast.Starred):
                    return i
            return -1
        star_idx = find_star(n.elts)
        if star_idx is -1:
            elts = list(map(self.visit, n.elts))
            return self.case_comp.seq_n(type, elts)

        ast_elts = n.elts
        elts1 = list(map(self.visit, ast_elts[:star_idx]))
        star = self.visit(ast_elts[star_idx].value)
        elts2 = list(map(self.visit, ast_elts[star_idx+1:]))
        return self.case_comp.seq_m_star_n(type, elts1, star, elts2)


    def visit_Tuple(self, n: ast.Tuple):
        return self._visit_seq(tuple, n)

    def visit_List(self, n: ast.List):
        return self._visit_seq(list, n)

    def visit(self, node):
        """Visit a node."""
        if isinstance(node, ConsistentConstant):
            return self.case_comp.literal(node)
        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, None)
        if visitor is None:
            raise TypeError(node)
        return visitor(node)


class GenMatch(ast.NodeTransformer):
    def __init__(self, token: str, activation):
        self.token = token
        self.activation = activation

        def id_gen():
            i = 0
            while True:
                yield "PM%d.%d" % (id(self.activation), i)
                i += 1

        self.local_id_generator = id_gen()

    @property
    def next_id(self):
        return next(self.local_id_generator)

    def visit_With(self, node: ast.With):
        if node.lineno not in self.activation:
            return self.generic_visit(node)

        if not len(node.items):
            return self.generic_visit(node)

        item = node.items[0].context_expr
        if not isinstance(item, ast.Call):
            return self.generic_visit(node)

        fn = item.func
        if not isinstance(fn, ast.Name) or fn.id != self.token:
            return self.generic_visit(node)

        assert not item.keywords

        assert all(isinstance(stmt, ast.If) for stmt in node.body)

        if len(item.args) is not 1:
            val_to_match = ast.Tuple(item.args, ast.Load())
        else:
            val_to_match = item.args[0]

        cached = Symbol(self.next_id, node.lineno, node.col_offset).to_name()

        ifs = node.body  # type: t.List[ast.If]
        for if_ in ifs:
            assert not if_.orelse

        case_comp = CaseCompilation()
        spb = SyntacticPatternBinding(case_comp)

        pairs = []
        for if_ in ifs:
            case = spb.visit(if_.test)
            stmts = Stmts(if_.body)
            pairs.append((case, stmts))

        res = case_comp.match(pairs)(Expr(cached))
        suite = res.suite
        suite.reverse()
        suite.append(ast.Assign([cached], val_to_match))
        suite.reverse()
        return suite

class PatternMatching(Extension):
    identifier = 'pattern-matching'

    def pre_rewrite_src(self, io: StringIO):
        io.write('from {} import {}\n'.format(__name__,
                                              NotExhaustive.__name__))

    def rewrite_ast(self, node: ast.AST):
        node = GenMatch(self.token, self.activation).visit(node)
        ExprContextFixer().visit(node)
        return node

    def __init__(self, token='match'):
        self.token = token



import moshmosh.extension_register
import unittest
import sys

class Test(unittest.TestCase):
    def test_extensions(self):
        import tests.lazy_import
        import tests.match
        import tests.template_python
        import tests.pipelines
        import tests.quick_lambdas
        import tests.scoped_operators
        if sys.version_info < (3, 5):
            import tests.complex_match_py34
        else:
            import tests.complex_match
        import tests.list_view
        import tests.incremental


# +lazy-import
from moshmosh.extensions import pattern_matching, scoped_operators, pipelines, template_python
# -lazy-import
from moshmosh.extension import Extension
from moshmosh.ast_compat import ast
import re

class LambdaCollector(ast.NodeTransformer):
    def __init__(self, pattern: 're.Pattern', mk_arg): # re.Pattern might not found
        self.mk_arg = mk_arg
        self.pattern = pattern
        self.max_arg_index = -1

    def found_quick_lambda(self):
        return self.max_arg_index is not -1

    def visit_Name(self, n: ast.Name):
        match = self.pattern.match(n.id)
        if match:
            ith = match.group('ith')
            if ith:
                ith = int(ith)
                self.max_arg_index = max(self.max_arg_index, ith)
                n.id = '.' + n.id
            else:
                self.max_arg_index = max(0, self.max_arg_index)
                n.id = '.' + self.mk_arg(0)

        return n

class CollectorDelegate:
    def __init__(self, cls, token, mk_arg):
        pattern = re.compile(mk_arg("(?P<ith>\d*)") + '$')
        self.collector = cls(pattern, mk_arg)
        self._mk_new = lambda : cls(pattern, mk_arg)

    def mk_new_(self):
        self.collector = self._mk_new()

    def get(self):
        return self.collector

class QuickLambdaDetector(ast.NodeTransformer):
    """
    scala-style lambdas, not recursively processed.
    """
    def __init__(self, activation, token: str):
        self.arg_col = CollectorDelegate(
            LambdaCollector,
            token,
            lambda x: "{}{}".format(token, x)
        )

        self.placeholder_col = CollectorDelegate(
            LambdaCollector,
            token,
            lambda x: "{}{}_".format(token, x)
        )
        self.activation = activation

    def visit_Call(self, call: ast.Call):
        if call.lineno in self.activation:
            def mk_quick_lam(col: CollectorDelegate, arg: ast.AST):
                if arg.lineno in self.activation:
                    assert not isinstance(arg, ast.Starred)
                    cur_collector = col.get()
                    arg = cur_collector.visit(arg)
                    if cur_collector.found_quick_lambda():
                        col.mk_new_()
                        argcount = cur_collector.max_arg_index + 1
                        return ast.Lambda(
                            ast.arguments(
                                [
                                    ast.arg(
                                        '.' + cur_collector.mk_arg(i),
                                        annotation=None
                                    )
                                    for i in range(argcount)
                                ],
                                vararg=None,
                                kwonlyargs=[],
                                kw_defaults=[],
                                kwarg=None,
                                defaults=[]
                            ),
                            arg
                        )
                return arg
            call.func = self.visit(call.func)
            call.args = [mk_quick_lam(self.arg_col, arg) for arg in call.args]
            call.keywords = [mk_quick_lam(self.arg_col, arg) for arg in call.keywords]
            return mk_quick_lam(self.placeholder_col, call)
        return  self.generic_visit(call)

class QuickLambda(Extension):
    identifier = "quick-lambda"
    def __init__(self, token: str=None):
        token = token or '_'
        self.token = token
        self.detector = QuickLambdaDetector(self.activation, token)

    def rewrite_ast(self, node):
        node = self.detector.visit(node)
        # from rbnf_rts.unparse import Unparser
        # Unparser(node)
        return node

    def __gt__(self, other):
        return isinstance(other, pattern_matching.PatternMatching)

    def __lt__(self, other):
        return isinstance(
            other,
            (
                template_python.Template,
                pipelines.Pipeline,
                scoped_operators.ScopedOperator
            )
        )



from importlib import import_module
from types import ModuleType

def import_module_plus(name, from_mod=None, package=None):
    if from_mod is not None:
        mod = import_module(from_mod, package)
        try:
            return getattr(mod, name)
        except AttributeError:
            path = mod.__file__
            exc = ImportError()
            exc.msg = "Module {} cannot import {}".format(
                from_mod,
                name
            )
            exc.path = path
            exc.name = from_mod
            raise exc
    else:
        return import_module(name, package)

def import_and_replace(globals, name, asname, from_mod, package):
    mod = import_module_plus(name, from_mod, package)
    if asname:
        globals[asname] = mod
    else:
        globals[name] = mod
    return mod

class LazyModule(ModuleType):
    __imp_info__ = ()

    def __init__(self, globals, name, asname=None, from_mod=None, package=None):
        object.__setattr__(self, '__imp_info__', (globals, name, asname, from_mod, package))
        if asname:
            globals[asname] = self
        else:
            globals[name] = self

    def __getattr__(self, name):
        imp_info = object.__getattribute__(self, '__imp_info__')
        mod = import_and_replace(*imp_info)
        return getattr(mod, name)

    def __delattr__(self, name):
        imp_info = object.__getattribute__(self, '__imp_info__')
        mod = import_and_replace(*imp_info)
        return delattr(mod, name)

    def __setattr__(self, name, value):
        imp_info = object.__getattribute__(self, '__imp_info__')
        mod = import_and_replace(*imp_info)
        return setattr(mod, name, value)



from moshmosh.ast_compat import ast
from moshmosh.extension import Extension, Activation
from moshmosh.rewrite_helper import ast_to_literal_without_locations
from moshmosh.ctx_fix import ExprContextFixer

runtime_ast_build = '_ast_build'
runtime_ast_copy = '_ast_copy'
_fix_ast_ctx = ExprContextFixer().visit

__all__ = ['build_ast']


def build_ast(d):
    nodes = literal_build_ast(d)
    nodes = fix_ast_ctx(nodes)

    mock = ast.Module(nodes)

    return mock.body


def fix_ast_ctx(node):
    if isinstance(node, list):
        for each in node:
            fix_ast_ctx(each)
    else:
        _fix_ast_ctx(node)
    return node


def literal_build_ast(literal):
    """
    Convert a python literal to an AST.
    """
    if isinstance(literal, dict):
        ctor = literal.pop('constructor')
        ctor = getattr(ast, ctor)
        return ctor(**{k: literal_build_ast(v) for k, v in literal.items()})

    if isinstance(literal, list):
        res = []
        for each in literal:
            each = literal_build_ast(each)
            if isinstance(each, list):
                res.extend(each)
            elif isinstance(each, ast.Expr) and isinstance(each.value, list):
                res.extend(each.value)
            else:
                res.append(each)
        return res

    return literal


class Symbol:
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return self.id

    def __iter__(self):
        yield self


class Splicing(ast.NodeTransformer):
    def __init__(self, syms):
        self.syms = {
            k: Symbol('{}({})'.format(runtime_ast_copy, k))
            for k in syms
        }

    def visit_Name(self, n: ast.Name):
        return self.syms.get(n.id, n)


class ArgumentCollector(ast.NodeVisitor):
    def __init__(self):
        self.args = set()

    def visit_arg(self, arg: ast.arg):
        self.args.add(arg.arg)


class MacroTransform(ast.NodeTransformer):
    def __init__(self, activation: Activation, token: str):
        self.activation = activation
        self.token = token

    def visit_FunctionDef(self, fn: ast.FunctionDef):
        if fn.lineno not in self.activation:
            return self.generic_visit(fn)

        if len(fn.decorator_list) is not 1:
            return self.generic_visit(fn)

        deco = fn.decorator_list[0]

        if not (isinstance(deco, ast.Name) and deco.id == self.token):
            return self.generic_visit(fn)

        assert fn.body
        fn.decorator_list.pop()
        arg_collector = ArgumentCollector()
        arg_collector.visit(fn)
        args = arg_collector.args
        splicing = Splicing(args)

        body_head = fn.body[-1]
        lineno, col_offset = body_head.lineno, body_head.col_offset
        new_body = []
        for stmt in fn.body:
            stmt = splicing.visit(stmt)
            new_body.append(stmt)

        mod = ast.parse(repr(ast_to_literal_without_locations(new_body)))  # type: ast.Module
        ast.fix_missing_locations(mod)
        expr = mod.body[0]  # type: ast.Expr
        value = expr.value
        fn.body = [
            ast.Return(
                ast.Call(ast.Name(runtime_ast_build, ast.Load()), [value], []),
                lineno=lineno,
                col_offset=col_offset)
        ]
        return fn


class Template(Extension):
    identifier = "template-python"

    def __init__(self, token="quote"):
        self.token = token
        self.visitor = MacroTransform(self.activation, self.token)
        self.ctx_fixer = ExprContextFixer()

    def pre_rewrite_src(self, io):
        io.write('from {} import build_ast as {}\n'.format(
            __name__, runtime_ast_build))
        io.write('from copy import deepcopy as {}\n'.format(runtime_ast_copy))

    def rewrite_ast(self, node: ast.AST):
        node = self.visitor.visit(node)
        self.ctx_fixer.visit(node)
        return node



from moshmosh.ast_compat import ast


class ExprContextWriter(ast.NodeVisitor):
    def __init__(self, ctx):
        self.ctx = ctx

    def _store_simply(self, node):
        node.ctx = self.ctx

    def _store_recursively(self, node):
        node.ctx = self.ctx
        self.generic_visit(node)

    visit_Name = _store_simply
    visit_Subscript = _store_simply
    visit_Attribute = _store_simply
    visit_Tuple = _store_recursively
    visit_List = _store_recursively
    visit_Starred = _store_recursively


_store_writer = ExprContextWriter(ast.Store())
_del_writer = ExprContextWriter(ast.Del())


class ExprContextFixer(ast.NodeVisitor):
    def visit_AnnAssign(self, n: ast.AnnAssign):
        _store_writer.visit(n.target)
        self.generic_visit(n.annotation)
        if n.value:
            self.generic_visit(n.value)

    def visit_AugAssign(self, n: ast.AugAssign):
        _store_writer.visit(n.target)
        self.generic_visit(n.value)

    def visit_Assign(self, n: ast.Assign):
        for target in n.targets:
            _store_writer.visit(target)
        self.generic_visit(n.value)

    def visit_Delete(self, n: ast.Delete):
        for target in n.targets:
            _del_writer.visit(target)

    def visit_Name(self, n: ast.Name):
        if not hasattr(n, 'ctx') or n.ctx is None:
            n.ctx = ast.Load()

    def visit_For(self, n: ast.For):
        _store_writer.visit(n.target)
        self.generic_visit(n)



# moshmosh?
# +template-python
import typing as t
from moshmosh.ast_compat import ast, get_constant
from moshmosh.extensions.pattern_matching.runtime import NotExhaustive
from toolz import compose
T = t.TypeVar('T')
G = t.TypeVar('G')
H = t.TypeVar('H')

not_exhaustive_err_type = ast.Name(NotExhaustive.__name__, ast.Load())


def quote(_):
    raise NotImplemented


class Names(dict):
    def __missing__(self, key):
        v = self[key] = len(self)
        return v


class Gensym:
    names = Names()

    def __init__(self, base_name):
        self.base = base_name

    def gen(self) -> ast.Name:
        i = self.names[self.base]
        self.names[self.base] += 1
        return ast.Name('{}.{}'.format(self.base, i), ast.Load())


class Expr(t.Generic[T]):
    value = None  # type: ast.expr

    def __init__(self, mk: ast.expr):
        self.value = mk


class Stmts(t.Generic[T]):
    suite = None  # type: t.List[ast.stmt]

    def __init__(self, mk: t.List[ast.stmt]):
        self.suite = mk


class Symbol(t.Generic[T]):
    def __init__(self, name, lineno, col_offset):
        self.name = name
        self.lineno = lineno
        self.col_offset = col_offset

    def to_name(self):
        return ast.Name(
            self.name,
            ast.Load(),
            lineno=self.lineno,
            col_offset=self.col_offset)


class Pattern(t.Generic[T, G]):
    def __init__(self, f):
        self.f = f

    def apply(self, i: Expr[T], remain: Stmts[G]) -> Stmts[G]:
        return self.f(i, remain)


def dyn_check(f):
    def func(a, b):
        assert isinstance(a, Expr), (f, type(a))
        assert isinstance(b, Stmts), (f, type(b))
        c = f(a, b)
        assert isinstance(c, Stmts), (f, type(c))
        return c

    return func


class CaseCompilation(t.Generic[G]):
    def __init__(self, ret_sym: str = '.RET'):
        self.ret = ast.Name(ret_sym, ast.Load())

    def literal(self, v):
        # noinspection PyStatementEffect PyUnusedLocal
        @quote
        def quote_eq(ret, v, expr, stmts):
            if v == expr:
                stmts
            else:
                ret = None  # failed

        @dyn_check
        def pat(target: Expr, remain: Stmts):
            stmts = quote_eq(self.ret, v, target.value,
                             remain.suite)
            return Stmts(stmts)

        return Pattern(pat)

    def pin(self, s: Expr):
        # noinspection PyStatementEffect PyUnusedLocal
        @quote
        def quote_eq(ret, v, expr, stmts):
            if v == expr:
                stmts
            else:
                ret = None

        @dyn_check
        def pat(target: Expr, remain: Stmts):
            stmts = quote_eq(self.ret, s.value, target.value,
                             remain.suite)
            return Stmts(stmts)

        return Pattern(pat)

    def wildcard(_):
        @dyn_check
        def pat(_, remain):
            return remain

        return Pattern(pat)

    # noinspection PyStatementEffect
    def capture(_, sym: Symbol):
        lhs = sym.to_name()

        # noinspection PyStatementEffect PyUnusedLocal
        @quote
        def quote_capture(lhs, target, remain):
            lhs = target
            remain

        @dyn_check
        def pat(target: Expr, remain: Stmts):
            stmts = quote_capture(lhs, target.value, remain.suite)
            return Stmts(stmts)

        return Pattern(pat)

    def intersect(_, ps):
        @dyn_check
        def pat(target, body):
            for p in reversed(ps):
                body = p.apply(target, body)
            return body

        return Pattern(pat)

    def alternative(self, ps):
        # noinspection PyStatementEffect PyUnusedLocal
        @quote
        def quote_alt(now_stmts, then_stmts, ret):
            now_stmts
            if ret is None:
                then_stmts

        @dyn_check
        def pat(target, body):
            then_code = Stmts([ast.Pass()])
            for each in reversed(ps):
                now_code = each.apply(target, body)  # type: Stmts
                assert isinstance(now_code, Stmts)
                stmts = quote_alt(now_code.suite, then_code.suite,
                                  self.ret)
                then_code = Stmts(stmts)
            return then_code

        return Pattern(pat)

    def recog(self, maker, item):
        def pmk(elts: t.List[Pattern]):
            @dyn_check
            def pat(target, body):
                n = len(elts)
                last = body
                mid = Gensym("recog").gen()
                for i in reversed(range(n)):
                    sub_tag = item(target, i)  # type: Stmts
                    assert isinstance(sub_tag, Expr)
                    last = elts[i].apply(Expr(mid), last)  # type: : Stmts
                    last = Stmts(
                        [ast.Assign([mid], sub_tag.value), *last.suite])
                return last

            return maker(Pattern(pat))

        return pmk

    def recog2(self, ctor: Expr, elts: t.List[Pattern]):
        """
        for deconstructors
        """

        @dyn_check
        def pat(target: Expr, body: Stmts):
            mid = Gensym("decons").gen()
            n = len(elts)

            # noinspection PyStatementEffect PyUnusedLocal
            @quote
            def decons(mid, ret, ctor, n, tag, body):
                mid = ctor.__match__(n, tag)
                if mid is None:
                    ret = None
                else:
                    body

            inner = self.seq_n(tuple, elts)
            stmts = inner.apply(Expr(mid), body)
            # noinspection PyTypeChecker
            suite = decons(mid, self.ret, ctor.value,
                           ast.Constant(n), target.value, stmts.suite)
            return Stmts(suite)

        return Pattern(pat)

    def instance_of(self, ty):
        # noinspection PyStatementEffect PyUnusedLocal
        @quote
        def quote_tychk(ret, tag, type, stmts):
            if isinstance(tag, type):
                stmts
            else:
                ret = None

        @dyn_check
        def pat(target: Expr, remain: Stmts):
            stmts = quote_tychk(self.ret, target.value, ty.value,
                                remain.suite)
            return Stmts(stmts)

        return Pattern(pat)

    def type_as(self, ty):
        if isinstance(ty, type):
            ty = ty.__name__
        ty = ast.Name(ty, ast.Load())

        def then(pattern):
            # noinspection PyStatementEffect PyUnusedLocal
            @quote
            def quote_tychk(ret, tag, ty, stmts):
                if isinstance(tag, ty):
                    stmts
                else:
                    ret = None

            @dyn_check
            def pat(target: Expr, remain: Stmts):
                remain = pattern.apply(target, remain)
                stmts = quote_tychk(self.ret, target.value, ty,
                                    remain.suite)
                return Stmts(stmts)

            return Pattern(pat)

        return then

    def size_is(self, n: int):
        n = ast.Constant(n)

        def then(pattern):
            if -5 < get_constant(n) < 256:
                # noinspection PyStatementEffect PyUnusedLocal
                @quote
                def quote_size_chk(ret, tag, n, stmts):
                    if len(tag) is n:
                        stmts
                    else:
                        ret = None
            else:
                # noinspection PyStatementEffect PyUnusedLocal
                @quote
                def quote_size_chk(ret, tag, n, stmts):
                    if len(tag) == n:
                        stmts
                    else:
                        ret = None

            @dyn_check
            def pat(target: Expr, remain: Stmts):
                remain = pattern.apply(target, remain)
                # noinspection PyTypeChecker
                stmts = quote_size_chk(self.ret, target.value, n,
                                       remain.suite)
                return Stmts(stmts)

            return Pattern(pat)

        return then

    def size_ge(self, n: int):
        n = ast.Constant(n)

        def then(pattern):
            @quote
            def quote_size_chk(ret, tag, n, stmts):
                if len(tag) >= n:
                    stmts
                else:
                    ret = None

            @dyn_check
            def pat(target: Expr, remain: Stmts):
                remain = pattern.apply(target, remain)
                # noinspection PyTypeChecker
                stmts = quote_size_chk(self.ret, target.value, n,
                                       remain.suite)
                return Stmts(stmts)

            return Pattern(pat)

        return then

    def seq_n(self, type, elts):
        return self.recog(
            compose(self.type_as(type), self.size_is(len(elts))),
            self.item)(elts)

    def seq_m_star_n(self, type, elts1, star, elts2):
        n1 = len(elts1)
        n2 = len(elts2)
        n = n1 + n2
        if n2 is 0:
            # when elts2 is empty,
            # use a[-1:None] instead of a[-1:-0]
            end = None
        else:
            end = -n2

        def item(expr: Expr, i):
            if i < n1:
                return self.item(expr, i)
            if i > n1:
                return self.item(expr, i - n)

            @quote
            def get_item(value, start, end):
                # noinspection PyStatementEffect
                value[start:end]

            # noinspection PyTypeChecker
            expr = get_item(expr.value, ast.Constant(n1), ast.Constant(end))[0]  # type: ast.Expr
            return Expr(expr.value)

        return self.recog(
            compose(self.type_as(type), self.size_ge(n)),
            item
        )(elts1 + [star] + elts2)

    def item(self, expr: Expr, i):
        @quote
        def get_item(value, ith):
            # noinspection PyStatementEffect
            value[ith]

        # noinspection PyTypeChecker
        expr = get_item(expr.value, ast.Constant(i))[0]  # type: ast.Expr
        return Expr(expr.value)

    def match(self, pairs):
        @quote
        def quote_alt(now_stmts, then_stmts, ret):
            now_stmts
            if ret is None:
                then_stmts

        @quote
        def quote_match_failed(not_exhaustive_err_type):
            raise not_exhaustive_err_type

        def pat(target):
            then_code = Stmts(quote_match_failed(not_exhaustive_err_type))
            for each, body in reversed(pairs):

                suite = body.suite
                suite.reverse()
                suite.append(ast.Assign([self.ret], ast.Constant(())))
                suite.reverse()

                now_code = each.apply(target, body)  # type: Stmts
                assert isinstance(now_code, Stmts)
                stmts = quote_alt(now_code.suite, then_code.suite,
                                  self.ret)
                then_code = Stmts(stmts)
            return then_code

        return pat



from moshmosh.extension import *
from moshmosh.extension import _extension_pragma_re_u

def update_pragmas(extension_builder: t.Dict[object, Extension], lines):
    """
    Traverse the source codes and extract out the scope of
    every extension. Incrementally.
    """
    # bind to local for faster visiting in the loop
    extension_pragma_re = _extension_pragma_re_u
    registered = Registered.extensions

    # for new cells of IPython, refresh the scoping info
    # of the extensions
    for ext in extension_builder.values():
        intervals = ext.activation.intervals
        if intervals:
            last = intervals.pop()
            intervals.clear()
            if type(last) is int:
                intervals.append(0)

    for i, line in enumerate(lines):
        pragma = extension_pragma_re.match(line)
        if pragma:
            pragma = pragma.groupdict()
            action = pragma['action']
            extension = pragma['ext']
            params = pragma['params'] or ''
            params = (param.strip() for param in params.split(','))
            params = tuple(i for i in params if i)
            try:
                ext_cls = registered[extension]
            except KeyError:
                # TODO: add source code position info
                raise ExtensionNotFoundError(extension)
            key = (ext_cls, params)

            ext = extension_builder.get(key, None)
            if ext is None:
                try:
                    ext = extension_builder[key] = ext_cls(*params)
                except Exception as e:
                    raise

            lineno = i + 1
            if action == "+":
                ext.activation.enable(lineno)
            else:
                ext.activation.disable(lineno)

    return list(extension_builder.values())

def perform_extension_incr(
    extension_builder: t.Dict[object, Extension],
    source_code,
    filename
):

    str_type = type(source_code)

    node = ast.parse(source_code, filename)
    if str_type is bytes:
        source_code = source_code.decode('utf8')

    extensions = update_pragmas(extension_builder, StringIO(source_code))
    extensions = sum(map(list, solve_deps(extensions)), [])

    string_io = StringIO()
    for each in extensions:
        each.pre_rewrite_src(string_io)

    for each in extensions:
        node = each.rewrite_ast(node)
        ast.fix_missing_locations(node)

    literal = ast_to_literal(node)
    string_io.write("import ast as _ast\n")
    string_io.write("from moshmosh.rewrite_helper import literal_to_ast as _literal_to_ast\n")
    string_io.write('\n')
    string_io.write('__literal__ = ')
    string_io.write(repr(literal))
    string_io.write('\n')
    string_io.write("__ast__ = _literal_to_ast(__literal__)\n")

    string_io.write('__code__ = compile')
    string_io.write('(__ast__, ')
    string_io.write('__import__("os").path.abspath("") if __name__ == "__main__" else __file__,')
    string_io.write('"exec")\n')

    string_io.write('exec(__code__, globals())\n')

    for each in extensions:
        each.post_rewrite_src(string_io)

    code = string_io.getvalue()
    if str_type is bytes:
        code = bytes(code, encoding='utf8')
    return code


class IPythonSupport:
    def __init__(self, builder: t.Dict[object, Extension]):
        self.builder = builder
        self.tmp_exts = None

    def input_transform(self, lines):
        self.tmp_exts = update_pragmas(self.builder, lines)
        return lines

    def ast_transform(self, node):
        extensions = self.tmp_exts
        extensions = sum(map(list, solve_deps(extensions)), [])

        prev_io = StringIO()
        for each in extensions:
            each.pre_rewrite_src(prev_io)

        prev_io.write("import ast as _ast\n")
        prev_io.write("from moshmosh.rewrite_helper import literal_to_ast as _literal_to_ast\n")
        prev_stmts = ast.parse(prev_io.getvalue()).body

        for each in extensions:
            node = each.rewrite_ast(node)
            ast.fix_missing_locations(node)

        post_io = StringIO()
        for each in extensions:
            each.post_rewrite_src(post_io)

        post_stmts = ast.parse(post_io.getvalue()).body
        node.body = prev_stmts + node.body + post_stmts
        return node



# moshmosh?
# +template-python

@quote
def f(x):
    x + 1
    x = y + 1

from moshmosh.ast_compat import ast
from astpretty import pprint

stmts = f(ast.Name("a"))

pprint(ast.fix_missing_locations(stmts[0]))
pprint(ast.fix_missing_locations(stmts[1]))



# moshmosh?
# +quick-lambda
# +pipeline
from tests.py35_unparse import Unparser
Unparser(__ast__)
from functools import reduce
print()
print()
x = map(_ + 1, _0_)
assert isinstance(x([1, 2, 3]), map)

x = [1, 2, 3] | map(_ + 1, _0_) | list
assert x == [2, 3, 4]

x = [3, 4, 5] | map(_ % 3, _0_) | list
assert x == [0, 1, 2]

assert reduce(_0 + _1, [10, 41, 59]) == 110

assert map(_1_, _0_)([1, 4, 3], _ + 2) | list == [3, 6, 5]



from moshmosh.extension import Extension
from moshmosh.ast_compat import ast

# https://github.com/python/cpython/blob/master/Parser/Python.asdl#L102
opname_map = {
    "+": 'Add',
    '-': 'Sub',
    '*': 'Mult',
    '/': 'Div',
    '%': 'Mod',
    '**': 'Pow',
    '<<': 'LShift',
    '>>': 'RShift',
    '|': 'BitOr',
    '^': 'BitXor',
    '&': 'BitAnd',
    '//': 'FloorDiv'
}


class ScopedOperatorVisitor(ast.NodeTransformer):
    """
    `a op b -> func(a, b)`, recursively.
    The `op => func` pair is specified by users.
    """
    def __init__(self, activation, op_name: str, func_name: str):
        self.pair = (op_name, func_name)
        self.activation = activation

    def visit_BinOp(self, n: ast.BinOp):
        if n.lineno in self.activation:
            name = n.op.__class__.__name__
            pair = self.pair
            if name == pair[0]:
                fn = ast.Name(pair[1], ast.Load())
                return ast.Call(
                    fn,
                    [self.visit(n.left), self.visit(n.right)],
                    [],
                    lineno=n.lineno,
                    col_offset=n.col_offset
                )

        return self.generic_visit(n)


class ScopedOperator(Extension):
    identifier = "scoped-operator"

    def __init__(self, op_name: str, func_name: str):
        self.op_name = opname_map.get(op_name, op_name)
        self.func_name = func_name
        self.visitor = ScopedOperatorVisitor(
            self.activation,
            self.op_name,
            self.func_name
        )

    def rewrite_ast(self, node):
        return self.visitor.visit(node)



# moshmosh?

from moshmosh.extensions.lazy_import import LazyModule
# +lazy-import
import io

assert issubclass(type(io), LazyModule)
assert isinstance(io, LazyModule)

repr(io)

assert not issubclass(type(io), LazyModule)
assert not isinstance(io, LazyModule)
# -lazy-import
import subprocess
assert not issubclass(type(subprocess), LazyModule)
assert not isinstance(subprocess, LazyModule)



from copy import copy
import itertools


class NotExhaustive(Exception):
    pass

class ListViewProspectiveGrowError(Exception):
    pass


class ListView(list):
    # noinspection PyMissingConstructor
    def __init__(self, src, view_indices):
        self.src = src
        self.view_indices = view_indices

    def count(self):
        return len(self.view_indices)

    def __getitem__(self, item):
        return self.src[self.view_indices[item]]

    def __setitem__(self, key, value):
        self.src[self.view_indices[key]] = value

    def __iter__(self):
        src = self.src
        for each in self.view_indices:
            yield src[each]

    def __contains__(self, item):
        return item in iter(self)

    def __add__(self, other):
        return list(self) + other

    def copy(self):
        return ListView(self.src, self.view_indices)

    def append(self, _):
        raise ListViewProspectiveGrowError

    def extend(self, _):
        raise ListViewProspectiveGrowError

    def pop(self, _=...):
        raise ListViewProspectiveGrowError

    def clear(self):
        raise ListViewProspectiveGrowError

    def remove(self, obj):
        raise ListViewProspectiveGrowError

    def reverse(self):
        view_indices = self.view_indices
        for i in range(len(view_indices) // 2):
            view_indices[i], view_indices[-i] = view_indices[-i], view_indices[
                i]

    def insert(self, _, __):
        raise ListViewProspectiveGrowError

    def index(self, obj, start=0, stop=None):
        src = self.src
        view_indices = self.view_indices
        n = len(view_indices)
        if stop is None:
            stop = n
        else:
            stop = min(n, stop)

        for i in range(start, stop):
            if src[view_indices[i]] == obj:
                return i
        raise ValueError

    def sort(self, *, key=None, reverse=False):
        src = self.src
        if key:

            def key_(i):
                return key(src[i])
        else:

            def key_(i):
                return src[i]

        if isinstance(self.view_indices, list):
            self.view_indices.sort(key=key_, reverse=reverse)
        else:
            self.view_indices = sorted(
                self.view_indices, key=key_, reverse=reverse)

    def __eq__(self, other):
        xs_l = iter(self)
        xs_r = iter(other)
        for l, r in zip(xs_l, xs_r):
            if l == r:
                continue
            return False
        try:
            next(xs_r)
        except StopIteration:
            return True
        return False

    def __repr__(self):
        return 'View{}'.format(list(self))



# moshmosh?
# +pattern-matching
from rbnf_rts.unparse import Unparser
Unparser(__ast__)
with match(2, 1, 2, 3):
    if (a, b, c) or [a, b, c]:
        print(a + b + c)
    if (hd, *tl):
        print(hd, tl)





from ast import *
a = Module(
    body=[
        Assign(
            lineno=1,
            col_offset=0,
            targets=[Name(lineno=4, col_offset=0, id='PM140085921707568.0', ctx=Store())],
            value=Tuple(
                lineno=1,
                col_offset=0,
                elts=[
                    Num(lineno=4, col_offset=11, n=1),
                    Num(lineno=4, col_offset=14, n=2),
                ],
                ctx=Load(),
            ),
        ),
        Expr(
            lineno=314,
            col_offset=12,
            value=[
                If(
                    lineno=242,
                    col_offset=16,
                    test=Call(
                        lineno=242,
                        col_offset=19,
                        func=Name(lineno=242, col_offset=19, id='isinstance', ctx=Load()),
                        args=[
                            Name(lineno=4, col_offset=0, id='PM140085921707568.0', ctx=Load()),
                            Name(lineno=242, col_offset=19, id='tuple', ctx=Load()),
                        ],
                        keywords=[],
                    ),
                    body=[
                        Expr(
                            lineno=243,
                            col_offset=20,
                            value=[

                            ],
                        ),
                    ],

                ),
            ],
        ),

    ],
)

compile(a, "x", "exec")


# moshmosh?
# +pattern-matching
# +quick-lambda
# +pipeline

class C:
    @classmethod
    def __match__(cls, i, x):
        if i is not 2:
            return None
        return x, x

def f1(x):
    with match(x):
        if C(C(a, b), C(c, d)): return (a, b, c, d)

assert f1(1) == (1, ) * 4

def f2(x, r=1):
    with match(x):
        if 0: return r
        if x: return f2(x-1, r * x)

assert f2(5) == 120

def f3(x):
# +pattern-matching
    with match(x):
        if [1, *x]:
            print(x)
            return sum(x)
        if (1, x, y): return x + y
        if _: return 0


assert f3([1, 2, 3, 4]) == 9
assert f3((1, 2, 3)) == 5
assert f3((1, 2, 3, 4)) == 0
assert f3(1) == 0

with match([4, 2, 3]):
    if [hd, *tl]:
        res = map(_ + hd, tl) | list

assert res == [6, 7]

with match(2, 1, 2, 3):
    if (a, b, c) or [a, b, c]:
        res = (a + b + c)
    if (hd, *tl):
        res = (hd, tl)

assert res == (2, (1, 2, 3))


def test_fn(data):
    with match(data):
        if (e, isinstance(int) and count):
            res =  [e] * count
        if (a, b, _) or [_, a, b]:
            res = (a + b)
        if (hd, *tl) or [hd, *tl]:
            res = (hd, tl)
        if "42":
            res = 42
    return res

assert test_fn([1, 2, 3]) == 5
assert test_fn((1, 2, 3)) == 3
assert test_fn((1, 2)) == [1, 1]
assert test_fn((object, 3)) == [object, object, object]
assert test_fn((object, 3, 4, 5)) == (object, (3, 4, 5))
assert test_fn("42") == 42
from moshmosh.extensions.pattern_matching import NotExhaustive

try:
    test_fn(1)
except NotExhaustive:
    pass


from moshmosh.repl_apis import *

ext_builder = {}

src = perform_extension_incr(ext_builder, """
# +pattern-matching

with match(1):
    if 1: res = 114


# -pattern-matching
# +pipeline
""", "yoyoyo")

scope = {'__file__': __file__}
exec(src, scope)
assert scope['res'] == 114

src = perform_extension_incr(ext_builder, """
with match(1):
    if 1: res = 114
""", "yoyoyo")

res = None
try:
    exec(src, scope)
except NameError as e:
    res = e.args[0]

assert "name 'match' is not defined" == res



src = perform_extension_incr(ext_builder, """
[1] | print
""", "yoyoyo")

res = None
exec(src, scope)


import sys
from moshmosh.extension import perform_extension, check_if_use_moshmosh_sys
from importlib._bootstrap import ModuleSpec
from importlib._bootstrap_external import (PathFinder, FileLoader,
                                           SourceFileLoader,
                                           SourcelessFileLoader,
                                           ExtensionFileLoader)

__all__ = []


class ProxySourceFileLoader(SourceFileLoader):
    def __init__(self, loader: SourceFileLoader):
        SourceFileLoader.__init__(self, loader.name, loader.path)

    def get_data(self, path: str):
        data = SourceFileLoader.get_data(self, path)
        if check_if_use_moshmosh_sys(data):
            return perform_extension(data, self.path)
        return data


class ProxySourcelessLoader(SourcelessFileLoader):
    def __init__(self, loader: SourcelessFileLoader):
        SourcelessFileLoader.__init__(self, loader.name, loader.path)

    def get_data(self, path: str):
        data = SourcelessFileLoader.get_data(self, path)
        if check_if_use_moshmosh_sys(data):
            return perform_extension(data, self.path)
        return data


class MoshmoshFinder(PathFinder):
    @classmethod
    def find_spec(cls, fullname, path=None, target=None):
        spec = PathFinder.find_spec(fullname, path, target)
        if spec and spec.loader and isinstance(spec.loader, FileLoader):
            loader = spec.loader
            loader_ty = loader.__class__
            loader_ = {
                SourcelessFileLoader: lambda: ProxySourcelessLoader(loader),
                SourceFileLoader: lambda: ProxySourceFileLoader(loader),
                ExtensionFileLoader: lambda: loader
            }[loader_ty]()
            spec.loader = loader_
        return spec


sys.meta_path.insert(0, MoshmoshFinder)



from astpretty import pprint
from ast import parse
# from moshmosh.extensions import pattern_matching
import ast
print_ast = ast.Name("print", ast.Load())


pprint(parse("""
from a import b
from . import c
import b
from . import *
"""))



from setuptools import setup, find_packages

with open('README.md') as readme:
    readme = readme.read()

version = "0.3"

setup(
    name='moshmosh-base',
    version=version if isinstance(version, str) else str(version),
    keywords="syntax, semantics, extension, macro, pattern matching", # keywords of your project separated by comma ","
    description="advanced syntax&semantics extension system for Python", # a concise introduction of your project
    long_description=readme,
    long_description_content_type="text/markdown",
    license='mit',
    url='https://github.com/thautwarm/moshmosh',
    author='thautwarm',
    author_email='twshere@outlook.com',
    packages=find_packages(),
    python_requires='>=3.5',
    entry_points={"console_scripts": []},
    # above option specifies commands to be installed,
    # e.g: entry_points={"console_scripts": ["yapypy=yapypy.cmd.compiler"]}
    install_requires=["toolz"],
    platforms="any",
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    zip_safe=False,
)


from moshmosh.extensions.pattern_matching import *
class MatchError(Exception):
    pass

class C:
    @classmethod
    def __match__(cls, i, x):
        if i is not 2:
            return None
        return x, x

def f1(x):
    with match(x):
        if C(C(a, b), C(c, d)): return (a, b, c, d)

assert f1(1) == (1, ) * 4

def f2(x, r=1):
    with match(x):
        if 0: return r
        if x: return f2(x-1, r * x)

assert f2(5) == 120

def f3(x):
    with match(x):
        if [1, x]:
            print(x)
            return sum(x)
        if (1, x, y): return x + y
        if _: return 0

assert f3([1, 2, 3, 4]) == 9

assert f3((1, 2, 3)) == 5
assert f3((1, 2, 3, 4)) == 0
assert f3(1) == 0

with match(2, 1):
    if (a, b, c) or [a, b, c]:
        res = (a + b + c)
    if (hd, tl):
        res = (hd, tl)

assert res == (2, 1)



# moshmosh?
# +pattern-matching

with match(1, 2):
    if (a, pin(3)):
        print(a)
    if (_, pin(2)) and (pin(1), _):
        print(10)
    if _:
        print(5)

with match([1, 2, 3, 4]):
    if [1, 2, a, b]:
        print(a + b)


class Succ:
    @classmethod
    def __match__(cls, n, tag):
        if n == 0 or not isinstance(tag, int):
            return None  # failed

        return tuple(tag + 1 for i in range(n))


def f(val):
    with match(val):
        if (4, Succ(3, x, y, z)) or (3, Succ(4, x, y, z)):
            print(x + y + z)
        if _:
            print('otherwise')


f((4, 2))
f((3, 3))


class GreaterThan:
    def __init__(self, v):
        self.v = v

    def __match__(self, cnt: int, to_match):
        if isinstance(to_match, int) and cnt is 0 and to_match > self.v:
            return ()  # matched


with match(114, 514):
    if (GreaterThan(42)() and a, b):
        print(b, a)



import ast


def ast_to_literal(node):
    """
    Convert an AST to a python literal.
    We avoid the use of comprehension expressions here
    to get more human-friendly call stacks.
    """
    if isinstance(node, ast.AST):
        field_names = node._fields
        res = {'constructor': node.__class__.__name__}
        for field_name in field_names:
            field = getattr(node, field_name, None)
            field = ast_to_literal(field)
            res[field_name] = field
        if hasattr(node, 'lineno'):
            res['lineno'] = node.lineno
        if hasattr(node, 'col_offset'):
            res['col_offset'] = node.col_offset

        return res
    if isinstance(node, list):
        res = []
        for each in node:
            res.append(ast_to_literal(each))
        return res
    return node

def ast_to_literal_without_locations(node):
    if isinstance(node, ast.AST):
        field_names = node._fields
        res = {'constructor': node.__class__.__name__}
        for field_name in field_names:
            field = getattr(node, field_name, None)
            field = ast_to_literal_without_locations(field)
            res[field_name] = field

        return res
    if isinstance(node, list):
        res = []
        for each in node:
            res.append(ast_to_literal_without_locations(each))
        return res
    return node

def literal_to_ast(literal):
    """
    Convert a python literal to an AST.
    """
    if isinstance(literal, dict):
        ctor = literal.pop('constructor')
        ctor = getattr(ast, ctor)

        return ctor(**{k: literal_to_ast(v) for k, v in literal.items()})

    if isinstance(literal, list):
        return list(map(literal_to_ast, literal))

    return literal



import abc
import typing as t
import re
import traceback
from io import StringIO
from moshmosh.rewrite_helper import ast_to_literal
from moshmosh.ast_compat import ast

_extension_token_b = re.compile(b"#\s*moshmosh\?\s*?")
_extension_token_u = re.compile(r"#\s*moshmosh\?\s*?")

_extension_pragma_re_u = re.compile(
    r'\s*#\s*(?P<action>[+-])(?P<ext>[^(\s]+)\s*(\((?P<params>.*)\))?[^\S\n]*?')

class Registered:
    extensions = {}  # type: t.Dict[str, t.Type[Extension]]

class Activation:
    """This sort of instances tell us
    whether an extension is enabled at a specific line number"""

    def __init__(self):
        self.intervals = []

    def enable(self, line: int):
        intervals = self.intervals
        if not intervals:
            intervals.append(line)
            return
        if isinstance(intervals[-1], int):
            # already enabled
            return

        intervals.append(line)

    def disable(self, line: int):
        intervals = self.intervals
        if not intervals or isinstance(intervals[-1], range):
            # already disabled
            return
        enable_line = intervals.pop()
        intervals.append(range(enable_line, line))

    def __contains__(self, item):
        for each in self.intervals:
            if isinstance(each, int):
                if item >= each:
                    return True
            else:
                assert isinstance(each, range)
                if item in each:
                    return True
        return False


class ExtensionNotFoundError(Exception):
    pass


class ExtensionMeta(type):
    def __new__(mcs, name, bases, ns: dict):
        if ns.get('_root', False):
            return super().__new__(mcs, name, bases, ns)

        bases = tuple(filter(lambda it: Extension is not it, bases))

        # All extensions need instantiating the its activation info.
        __init__ = ns.get('__init__', None)
        if __init__:
            def init(self, *args, **kwargs):
                self.activation = Activation()
                __init__(self, *args, **kwargs)
        else:
            def init(self):
                self.activation = Activation()
        if 'activation' not in ns:
            ns['activation'] = None

        assert 'identifier' in ns, "An extension should have its identifier."

        ns['__init__'] = init
        ns = {
            **Extension.__dict__,
            **ns
        }
        ret = type(name, bases, ns)  # type: t.Type[RealExtension]
        Registered.extensions[ret.identifier] = ret

        return ret

    def __instancecheck__(self, other):
        return other in Registered.extensions.values()

class Extension(metaclass=ExtensionMeta):
    """automatically extension"""

    _root = True

    @property
    def activation(self) -> Activation:
        raise NotImplemented

    @property
    @abc.abstractmethod
    def identifier(cls):
        "A string to indicate the class of extension instance."
        raise NotImplemented

    def pre_rewrite_src(self, io: StringIO):
        pass

    @abc.abstractmethod
    def rewrite_ast(self, node: ast.AST):
        "A function to perform AST level rewriting"
        raise NotImplemented

    def post_rewrite_src(self, io: StringIO):
        pass

    def __gt__(self, other):
        return False

    def __lt__(self, other):
        return False


class RequirementNotResolved(Exception):
    pass

def solve_deps(exts):
    deps = {ext: set() for ext in exts}
    # build dependencies
    for ext in exts:
        for other in exts:
            if ext is other:
                continue
            if ext > other or other <  ext:
                deps[ext].add(other)
    groups = []
    while deps:
        group = set.union(*deps.values()).difference(deps.keys())
        to_del_deps = []
        for k, v in deps.items():
            if not v:
                to_del_deps.append(k)
        group.update(to_del_deps)
        if not group:
            raise RequirementNotResolved

        for k in to_del_deps:
            del deps[k]

        for v in deps.values():
            v.difference_update(group)
        groups.append(group)
    return groups


def extract_pragmas(lines):
    """
    Traverse the source codes and extract out the scope of
    every extension.
    """
    # bind to local for faster visiting in the loop
    extension_pragma_re = _extension_pragma_re_u
    registered = Registered.extensions
    extension_builder = {}  # type: t.Dict[object, Extension]

    for i, line in enumerate(lines):
        pragma = extension_pragma_re.match(line)
        if pragma:
            pragma = pragma.groupdict()
            action = pragma['action']
            extension = pragma['ext']
            params = pragma['params'] or ''
            params = (param.strip() for param in params.split(','))
            params = tuple(i for i in params if i)
            try:
                ext_cls = registered[extension]
            except KeyError:
                # TODO: add source code position info
                raise ExtensionNotFoundError(extension)
            key = (ext_cls, params)

            ext = extension_builder.get(key, None)
            if ext is None:
                try:
                    ext = extension_builder[key] = ext_cls(*params)
                except Exception as e:
                    raise

            lineno = i + 1
            if action == "+":
                ext.activation.enable(lineno)
            else:
                ext.activation.disable(lineno)

    return list(extension_builder.values())


def _stack_exc(f):
    def apply(source_code, filename='<unknown>'):
        try:
            return f(source_code, filename)
        except Exception as e:
            traceback.print_tb(e.__traceback__)
            raise e

    return apply


def check_if_use_moshmosh_sys(source_code):
    str_type = type(source_code)
    extension_token = _extension_token_b if str_type is bytes else _extension_token_u
    return bool(extension_token.match(source_code))


@_stack_exc
def perform_extension(source_code, filename):

    str_type = type(source_code)

    node = ast.parse(source_code, filename)
    if str_type is bytes:
        source_code = source_code.decode('utf8')

    extensions = extract_pragmas(StringIO(source_code))
    extensions = sum(map(list, solve_deps(extensions)), [])

    string_io = StringIO()
    for each in extensions:
        each.pre_rewrite_src(string_io)

    for each in extensions:
        node = each.rewrite_ast(node)
        ast.fix_missing_locations(node)

    literal = ast_to_literal(node)
    string_io.write("import ast as _ast\n")
    string_io.write("from moshmosh.rewrite_helper import literal_to_ast as _literal_to_ast\n")
    string_io.write('\n')
    string_io.write('__literal__ = ')
    string_io.write(repr(literal))
    string_io.write('\n')
    string_io.write("__ast__ = _literal_to_ast(__literal__)\n")

    string_io.write('__code__ = compile')
    string_io.write('(__ast__, ')
    string_io.write('__import__("os").path.abspath("") if __name__ == "__main__" else __file__,')
    string_io.write('"exec")\n')

    string_io.write('exec(__code__, globals())\n')

    for each in extensions:
        each.post_rewrite_src(string_io)

    code = string_io.getvalue()
    if str_type is bytes:
        code = bytes(code, encoding='utf8')
    return code



from .ast_compat import ast
from .extension_register import *
from .extensions import template_python
from .extensions import lazy_import
from .extensions import pattern_matching
from .extensions import scoped_operators
from .extensions import pipelines
from .extensions import quick_lambdas
